
#include "MAG-CAS-plugin.h"

#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <time.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <sys/select.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include <net/if.h>

#include <pthread.h>
#include <openssl/bio.h>
#include <openssl/pem.h>
#include <openssl/conf.h>
#include <openssl/x509.h>
#include <openssl/ssl.h>
#include <openssl/err.h>
#include <openssl/md5.h>


#define CAS_ROOTCERT_FILE "/flash/cubitel.crt"
#define CAS_USERCERT_FILE "/mnt/Userfs/client.crt"
#define CAS_USERKEY_FILE "/mnt/Userfs/client.key"
#define CAS_REQUEST_FILE "/mnt/Userfs/client.req"
#define CAS_SERVER_HOST "<your server name here>"
#define CAS_SERVER_PORT "1515"
#define CAS_SYSLOG_SERVER "10.39.0.2"
#define CAS_KEYHASH "0000000000000000"


#define CAS_SYSID	0x1515
#define CAS_SOID	0x0000

#define CAS_CMD_PING	0
#define CAS_CMD_NEWPROG	1	/* 2 bytes pnum */
#define CAS_CMD_GETKEY	2	/* 2 bytes pnum, 2 bytes id */
#define CAS_CMD_STOP	3
#define CAS_CMD_KEY		4	/* 2 bytes pnum, 2 bytes id, 16 bytes key */

#define MAX_OPKEYS	8

#define SP_WRITE	0
#define SP_READ		1


/**
 * ECM Format
 *   0x00 01  0x00
 *   0x01 01  Table ID (80 / 81)
 *   0x03 02  Section Size
 *   0x04 01  CAS Version (01)
 *   0x05 01  Scrambling type
 *   0x06 02  Program ID
 *   0x08 02  Operational Key ID
 *   0x0A 06  <reserved>
 *   0x10 16  AES IV
 *   0x20 32  Encrypted keys
 *   0x40 04  Table CRC
 */


typedef struct opkey {
	unsigned short		pnum;
	unsigned short		id;
	unsigned char		key[16];
} opkey_t;

static STB_MAG_Cas_t PluginInterface;
static STB_MAG_Cad_t DescramblerInterface = {NULL,NULL};
static int sock;
static int sockpair[2];
static unsigned char mac[6];
static pthread_t thread;

/* ECM cache */
static unsigned char lastecm[188];
/* OpKeys cache */
static unsigned char opkeysFirst = 0;
static unsigned char opkeysLast = 0;
static opkey_t opkeys[MAX_OPKEYS];


static void casSendString(char *str)
{
	struct sockaddr_in sa;
	char buf[250];
	
	sprintf(buf, "%02x%02x%02x%02x%02x%02x: %s", mac[0], mac[1], mac[2], mac[3], mac[4], mac[5], str);

	sa.sin_family = AF_INET;
	sa.sin_addr.s_addr = inet_addr(CAS_SYSLOG_SERVER);
	sa.sin_port = htons(514);
	sendto(sock, buf, strlen(buf), 0, (struct sockaddr *)&sa, sizeof(sa));
}

static void casSendError()
{
	char buf[250];
	
	sprintf(buf, "SSL Error %08X.", ERR_get_error());
	casSendString(buf);
}


static int sslPasswordCallback(char *buf, int size, int rwflag, void *userdata)
{
	MD5_CTX md;
	int i;
	unsigned char hash[16];
	
	MD5_Init(&md);
	MD5_Update(&md, mac, 6);
	MD5_Update(&md, CAS_KEYHASH, 16);
	MD5_Final(hash, &md);
	
	for (i = 0; i < 16; i++) sprintf(&buf[i*2], "%02x", hash[i]);

	return 32;
}

static int sslRead(BIO *bio, unsigned char *buf, int len)
{
	int r, fd, nfds;
	fd_set rfds, rfd;
	struct timeval tv;
	unsigned char wrbuf[512];
	
	fd = BIO_get_fd(bio, NULL);
	
	FD_ZERO(&rfds);
	FD_SET(fd, &rfds);
	FD_SET(sockpair[SP_READ], &rfds);
	nfds = (fd > sockpair[SP_READ]) ? fd : sockpair[SP_READ];
	
	while (1) {
		tv.tv_sec = 10;
		tv.tv_usec = 0;
		r = select(nfds+1, &rfds, NULL, NULL, &tv);
		if (r <= 0) return r;
	
		if (FD_ISSET(fd, &rfds)) {
			r = BIO_read(bio, buf, len);
			/* EOF */
			if (r == 0) return -1;
			/* Data */
			if (r > 0) return r;

			while (errno == EAGAIN) {
				FD_ZERO(&rfd);
				FD_SET(fd, &rfd);
				tv.tv_sec = 2;
				tv.tv_usec = 0;
				r = select(fd+1, &rfd, NULL, NULL, &tv);
				if (r <= 0) return r;
				
				r = BIO_read(bio, buf, len);
				if (r >= 0) return r;
			}

			break;
		}
		
		if (FD_ISSET(sockpair[SP_READ], &rfds)) {
			r = recv(sockpair[SP_READ], wrbuf, sizeof(wrbuf), 0);
			if (r <= 0) return -1;
			
			BIO_write(bio, wrbuf, r);
		}
	}
	
	return -1;
}

static int sslWrite(unsigned char *buf, int len)
{
	return send(sockpair[SP_WRITE], buf, len, 0);
}

static void sslCreateCertRequest()
{
	FILE *fReq, *fKey;
	X509_REQ *req;
	EVP_PKEY *pk;
	RSA *rsa;
	X509_NAME *name;
	char cn[30];
	
	pk = EVP_PKEY_new();
	if (!pk) return;
	
	req = X509_REQ_new();
	if (!req) {
		EVP_PKEY_free(pk);
		return;
	}
	
	rsa = RSA_generate_key(1024, RSA_F4, NULL, NULL);
	EVP_PKEY_assign_RSA(pk, rsa);
	X509_REQ_set_pubkey(req, pk);
	
	name = X509_REQ_get_subject_name(req);
	sprintf(cn, "%02x:%02x:%02x:%02x:%02x:%02x", mac[0], mac[1], mac[2], mac[3], mac[4], mac[5]);
	X509_NAME_add_entry_by_txt(name, "CN", MBSTRING_ASC, cn, -1, -1, 0);
	X509_NAME_add_entry_by_txt(name, "O", MBSTRING_ASC, "Cubitel JSC", -1, -1, 0);
	X509_NAME_add_entry_by_txt(name, "C", MBSTRING_ASC, "RU", -1, -1, 0);
	
	X509_REQ_sign(req, pk, EVP_md5());
	
	fReq = fopen(CAS_REQUEST_FILE, "w+");
	if (!fReq) {
		return;
	}
	fKey = fopen(CAS_USERKEY_FILE, "w+");
	if (!fKey) {
		fclose(fReq);
		return;
	}
	
	PEM_write_X509_REQ(fReq, req);
	PEM_write_PKCS8PrivateKey(fKey, pk, EVP_aes_256_cbc(), NULL, 0, sslPasswordCallback, NULL);
	
	fclose(fReq);
	fclose(fKey);
	
	X509_REQ_free(req);
	EVP_PKEY_free(pk);
}

/**
 * Decrypt ECM and set working keys
 */
static void casECM(byte *ecm, opkey_t *key)
{
	EVP_CIPHER_CTX ctx;
	unsigned char keys[32];

	EVP_CIPHER_CTX_init(&ctx);

	do {
		if (EVP_EncryptInit_ex(&ctx, EVP_aes_128_ofb(), NULL, key->key, &ecm[0x10]) <= 0) break;

		EVP_Cipher(&ctx, keys, &ecm[0x20], 32);
	} while (0);

    EVP_CIPHER_CTX_cleanup(&ctx);

	if (DescramblerInterface.SetScramblingKey && DescramblerInterface.SetScramblingType) {
		DescramblerInterface.SetScramblingType(ecm[0x05]);
		DescramblerInterface.SetScramblingKey(&keys[0], 0);
		DescramblerInterface.SetScramblingKey(&keys[16], 1);
	}
}

static opkey_t *casFindKey(unsigned short pnum, unsigned short id)
{
	int i;

	/* Search opkey in cache */
	for (i = opkeysFirst; i == opkeysLast; (++i == MAX_OPKEYS) ? (i = 0) : i ) {
		if ( (opkeys[i].pnum == pnum) && (opkeys[i].id == id) ) return &opkeys[i];
	}
	
	return NULL;
}

static void casServerPacket(int cmd, unsigned char *buf, int len)
{
	unsigned short pnum, id;
	opkey_t *key;
	char s[100];
	
	if (cmd == CAS_CMD_KEY) {
		/* Check message size */
		if (len < sizeof(opkey_t)) return;
		key = (opkey_t *)buf;
		
		/* Find key in cache */
		if (casFindKey(key->pnum, key->id)) return;
		
		/* Store key in cache */
		memcpy(&opkeys[opkeysLast], key, sizeof(opkey_t));
		if (++opkeysLast == MAX_OPKEYS) opkeysLast = 0;
		if ( (opkeysLast == opkeysFirst) && (++opkeysFirst == MAX_OPKEYS) ) opkeysFirst = 0;

		pnum = (lastecm[6] << 8) | lastecm[7];
		id = (lastecm[8] << 8) | lastecm[9];
		if ( (key->pnum == pnum) && (key->id == id) ) {
			casECM(lastecm, key);
		}
	}
}

static void *casThread(void *p)
{
	BIO *bio;
	SSL_CTX *ctx;
	SSL *ssl;
	int r, len, cmd;
	unsigned char buf[512];
	
	socketpair(AF_UNIX, SOCK_DGRAM, 0, &sockpair[0]);
	
	while (1) {
	
		r = open(CAS_USERKEY_FILE, O_RDONLY);
		if (r >= 0) {
			close(r);
		} else {
			sslCreateCertRequest();
		}

		bio = NULL;
		ctx = NULL;
		ssl = NULL;

		do {
			/* Initialize SSL context */
			ctx = SSL_CTX_new(SSLv23_client_method());
			if (!ctx) {
				casSendError();
				break;
			}
			
			/* Load client and root certificates */
			SSL_CTX_load_verify_locations(ctx, CAS_ROOTCERT_FILE, NULL);
			SSL_CTX_set_default_passwd_cb(ctx, sslPasswordCallback);
			
			if (SSL_CTX_use_certificate_file(ctx, CAS_USERCERT_FILE, SSL_FILETYPE_PEM) < 1) {
				break;
			}
			if (SSL_CTX_use_PrivateKey_file(ctx, CAS_USERKEY_FILE, SSL_FILETYPE_PEM) < 1) {
				break;
			}
			
			bio = BIO_new_ssl_connect(ctx);
			if (!bio) break;
			BIO_get_ssl(bio, &ssl);
			SSL_set_mode(ssl, SSL_MODE_AUTO_RETRY);
			
			/* Connecting to server */
			BIO_set_conn_hostname(bio, CAS_SERVER_HOST ":" CAS_SERVER_PORT);
			if (BIO_do_connect(bio) <= 0) break;
			
			/* Verify server certificate */
//			if (SSL_get_verify_result(ssl) != X509_V_OK) {
//				casSendString("Bad CAS server certificate!");
//				break;
//			}
			
			/* Unblock socket */
			BIO_set_nbio(bio, 1);
			
			/* Read packets from server in loop */
			while (1) {
				/* Two bytes length and one byte command, then variable data */
				r = sslRead(bio, buf, sizeof(buf));
				if (r < 3) {
					if (r < 0) {
						break;
					}
					continue;
				}
				len = (buf[0] << 8) | buf[1];
				cmd = buf[2];
				if (len > r) {
					break;
				}

				casServerPacket(cmd, &buf[3], len);
			}
			
			BIO_reset(bio);
		} while (0);
		
		if (ctx)
			SSL_CTX_free(ctx);

		if (bio)
			BIO_free_all(bio);

		sleep(10);
	}

	return NULL;
}

/**
 * Callback to be called by Player on PMT change
 */
static void CAS_OnPmtChange(byte* buffer, int length)
{
	unsigned char pkt[5];
	
	pkt[0] = 0;
	pkt[1] = 2;
	pkt[2] = CAS_CMD_NEWPROG;
	pkt[3] = buffer[4];
	pkt[4] = buffer[5];
	sslWrite(pkt, 5);
	
	return;
};

/**
 * Check ECM packet, find operational key
 */
static void CAS_OnEcm(byte* ecm, int length)
{
	unsigned short pnum;
	unsigned short id;
	unsigned char pkt[7];
	opkey_t *key;
	
	/* Check for valid ECM */
	if (length < 16) return;
	if (ecm[4] != 1) return;

	/* Check if ECM is changed */
	if (!memcmp(ecm, lastecm, 16)) return;
	memcpy(lastecm, ecm, length);
	
	pnum = (ecm[6] << 8) | ecm[7];
	id = (ecm[8] << 8) | ecm[9];
	
	key = casFindKey(pnum , id);
	if (key) {
		casECM(ecm, key);
	} else {
		pkt[0] = 0;
		pkt[1] = 4;
		pkt[2] = CAS_CMD_GETKEY;
		pkt[3] = pnum >> 8;
		pkt[4] = pnum & 0xFF;
		pkt[5] = id >> 8;
		pkt[6] = id & 0xFF;
		sslWrite(pkt, 7);
	}

	return;
};

static int CAS_Decrypt(byte* in_buffer, int length)
{
	return length;
};

/**
 * Free up CA module
 */
static void CAS_Deinit(void)
{
	pthread_cancel(thread);
	pthread_join(thread, NULL);
	close(sock);

	return;
};

/**
 * Reset CAS context. Usually called to indicate that user has
 * switched to another stream
 */
static void CAS_ResetStream(void)
{
	unsigned char pkt[3];
	
	pkt[0] = 0;
	pkt[1] = 0;
	pkt[2] = CAS_CMD_STOP;
	sslWrite(pkt, 3);

	return;
};

/**
 * Get CAS vendor SysID as defined in DVB. Useful only in
 * CAS_Flags_PMT_ECM mode
 */
static int CAS_GetSysID(void)
{
	return CAS_SYSID;
};

/**
 * Get CAS vendor SoID as defined in DVB. Useful only in
 * CAS_Flags_PMT_ECM mode
 */
static int CAS_GetSoID(void)
{
	return CAS_SOID;
};

/**
 * Get CAS vendor specific flags
 *
 * @see CAS_Flags_e
 */
static CAS_Flags_e CAS_GetCasFlags(void)
{
	return CAS_Flags_PMT_ECM;
};

/**
 * Get the format of ECM data expected by the plugin in
 * CAS_OnEcm
 */
static ECM_Format_e CAS_GetEcmFormat(void)
{
  return ECM_FORMAT_SECTION;
};

/**
 * Set some custom param to CAS plugin.
 *
 * name - param name always converted to upper case
 * value - param value
 */
static void CAS_SetAdditionalParam(const char* name, const char* value)
{
};


/* dynamic library will implement this */
STB_MAG_Cas_t* CreateCasPlugin(STB_MAG_Cad_t* mag_interface, const char* ini_filename)
{
	struct sockaddr_in sa;
	struct ifreq ifr;
	
	/* Initialize OpenSSL */
	SSL_library_init();
	OpenSSL_add_all_algorithms();
	
	/* create debug socket */
	sock = socket(AF_INET, SOCK_DGRAM, 0);
	
	/* get STB MAC address */
	memset(&ifr, 0, sizeof(ifr));
	strcpy(ifr.ifr_name, "eth0");
	ioctl(sock, SIOCGIFHWADDR, &ifr);
	memcpy(mac, &ifr.ifr_hwaddr.sa_data, 6);

	/* Bind */
	sa.sin_family = AF_INET;
	sa.sin_addr.s_addr = 0;
	sa.sin_port = htons(5555);
	bind(sock, (struct sockaddr *)&sa, sizeof(sa));
	
	/* Test */
	casSendString("Startup OK");
	
	/* Create network thread */
	pthread_create(&thread, NULL, casThread, NULL);
	
	if(mag_interface) {
		DescramblerInterface.SetScramblingKey = mag_interface->SetScramblingKey;
		DescramblerInterface.SetScramblingType = mag_interface->SetScramblingType;
	}

	PluginInterface.Decrypt = CAS_Decrypt;
	PluginInterface.Deinit = CAS_Deinit;
	PluginInterface.GetCasFlags = CAS_GetCasFlags;
	PluginInterface.GetEcmFormat = CAS_GetEcmFormat;
	PluginInterface.GetSoID = CAS_GetSoID;
	PluginInterface.GetSysID = CAS_GetSysID;
	PluginInterface.OnEcm = CAS_OnEcm;
	PluginInterface.OnPmtChange = CAS_OnPmtChange;
	PluginInterface.ResetStream = CAS_ResetStream;
	PluginInterface.SetAdditionalParam = CAS_SetAdditionalParam;
	return &PluginInterface;
};

/**
 * CAS API implementation version of the CAS vendor library
 * dynamic library will implement this
 *
 * @see CAS_API_VERSION
 */
int GetCasApiVersion(void)
{
	return CAS_API_VERSION;
};

/**
 * Get description of CAS plugin (e.i. Vendor, ...)
 */
const char* GetCasPluginDescription(void)
{
	return "Cubitel CAS plugin";
};
